#include <cuex/algo.h>
#include <cuex/oprdefs.h>
#include <cuex/opn.h>
#include <cuex/var.h>
#include <cuex/recursion.h>
#include <cucon/frame.h>
#include <cu/test.h>
#include <cu/int.h>
#include <stdlib.h>

#define VERBOSE 0
#if 1
#  define REPEAT 100000
#  define PLOT_BIN_COUNT 20
#  define PLOT_NODECOUNT_GRAN 10
#else
#  define REPEAT 10
#  define PLOT_BIN_COUNT 10
#  define PLOT_NODECOUNT_GRAN 1
#endif
#define MAX_NODE_COUNT (PLOT_BIN_COUNT*PLOT_NODECOUNT_GRAN)

static cu_bool_t
dmu_depth_eq_aux(cuex_t e0, cuex_t *muf0,
		 cuex_t e1, cuex_t *muf1, int depth)
{
    cuex_meta_t m0, m1;

    if (depth < 0)
	return cu_true;

    m0 = cuex_meta(e0);
    m1 = cuex_meta(e1);
    if (cuex_is_rvarmeta(m0)) {
	int i = cuex_varmeta_index(m0);
	muf0 = cucon_frame_at(muf0, i);
	e0 = *muf0;
	m0 = cuex_oa_mubind_set(cuex_meta(e0), 0);
    }
    else if (cuex_meta_is_opr(m0) && cuex_oa_mubind(m0)) {
	muf0 = cucon_frame_push(muf0, sizeof(cuex_t));
	*muf0 = e0;
	m0 = cuex_oa_mubind_set(m0, 0);
    }

    if (cuex_is_rvarmeta(m1)) {
	int i = cuex_varmeta_index(m1);
	muf1 = cucon_frame_at(muf1, i);
	e1 = *muf1;
	m1 = cuex_oa_mubind_set(cuex_meta(e1), 0);
    }
    else if (cuex_meta_is_opr(m1) && cuex_oa_mubind(m1)) {
	muf1 = cucon_frame_push(muf1, sizeof(cuex_t));
	*muf1 = e1;
	m1 = cuex_oa_mubind_set(m1, 0);
    }

    if (m0 != m1)
	return cu_false;
    if (cuex_meta_is_opr(m0)) {
	cu_rank_t i, r;
	r = cuex_opr_r(m0);
	for (i = 0; i < r; ++i)
	    if (!dmu_depth_eq_aux(cuex_opn_at(e0, i), muf0,
				  cuex_opn_at(e1, i), muf1, depth - 1))
		return cu_false;
	return cu_true;
    }
    else
	return e0 == e1;
}

/* Approximation to the equal predicate for μ-recursive expressions. */
static cu_bool_t
dmu_depth_eq(cuex_t e0, cuex_t e1)
{
    int d;
    d = 2*cu_int_max(cuex_max_depth(e0), cuex_max_depth(e1));
    if (d > 10)
	d = 10; /* Avoid combinatorial explosion. */
    return dmu_depth_eq_aux(e0, NULL, e1, NULL, d);
}


cuex_t
random_expr(int depth, int *size_limit)
{
    cuex_t e_arr[2];
    cuex_meta_t opr;
    int i, r;
    switch (*size_limit > 1? lrand48() % (3 + !!depth) : 0) {
	case 0:
	    return cuex_var_new_e();
	case 1:
	    opr = CUEX_O1_SINGLETON;
	    break;
	case 2:
	    opr = CUEX_O2_GPROD;
	    break;
	case 3:
	    return cuex_rvar(lrand48() % depth);
    }
    --*size_limit;
    if (lrand48() % 2) {
	opr |= CUEX_OA_MUBIND(1);
	++depth;
    }
    r = cuex_opr_r(opr);
    for (i = 0; i < r; ++i)
	e_arr[i] = random_expr(depth, size_limit);
    return cuex_opn_by_arr(opr, e_arr);
}

cuex_t
unfold_random(cuex_t e, int *n, double x)
{
    cuex_meta_t e_meta = cuex_meta(e);
    if (*n > 0 && cuex_meta_is_opr(e_meta)) {
	if (cuex_oa_mubind(e_meta) && drand48() < x) {
	    e = cuex_mu_unfold(e);
	    --*n;
	}
	else {
	    x /= cuex_opr_r(e_meta);
	    CUEX_OPN_TRAN(e_meta, e, ep, unfold_random(ep, n, x));
	}
    }
    return e;
}

struct run_stats_s
{
    int count;
    clock_t time;
};

void
test(struct run_stats_s *stats_arr)
{
    static int run = 0;
    cuex_t e, ep, epp, eppp, eT;
    cu_bool_t are_eq;
    int size_limit = 10;
    int n, plot_bin;
    size_t eT_size, epp_size;
    cuex_stats_t stats;

    e = random_expr(0, &size_limit);
    ep = cuex_mu_minimise(e);
    are_eq = dmu_depth_eq(e, ep);
    if (!are_eq || (VERBOSE && e != ep && run++ < 30))
	cu_fprintf(stderr, "\n INPUT: %!\nOUTPUT: %!\n", e, ep);
    if (!are_eq)
	cu_test_bugf("Equality test failed.");

    n = 1;
    eT = unfold_random(ep, &n, 0.5);
    cuex_stats(eT, &stats);
    eT_size = stats.node_cnt + stats.var_cnt;
    do {
	n = 2;
	epp = eT;
	epp_size = eT_size;
	eT = unfold_random(eT, &n, 0.5);
	cuex_stats(eT, &stats);
	eT_size = stats.node_cnt + stats.var_cnt;
    } while (lrand48() % (MAX_NODE_COUNT/4) && eT_size < MAX_NODE_COUNT);
    if (epp_size >= MAX_NODE_COUNT)
	return;

    plot_bin = epp_size/PLOT_NODECOUNT_GRAN;
    stats_arr[plot_bin].time -= clock();
    eppp = cuex_mu_minimise(epp);
    stats_arr[plot_bin].time += clock();
    ++stats_arr[plot_bin].count;

    if (eppp != ep) {
	cu_fprintf(stderr, "   INPUT: %!\n  OUTPUT: %!\nEXPECTED: %!\n",
		   epp, eppp, ep);
	cu_test_bugf("Incomplete or incorrect simplification.");
    }
}

int
main()
{
    FILE *out;
    int i;
    struct run_stats_s stats_arr[PLOT_BIN_COUNT];
    cuex_init();
    cu_test_on_bug(cu_test_bugaction_cont);

    memset(stats_arr, 0, sizeof(stats_arr));
    for (i = 0; i < REPEAT; ++i)
	test(stats_arr);

    out = fopen("tmp.hopcroft_t0.data", "w");
    if (out) {
	for (i = 0; i < PLOT_BIN_COUNT; ++i)
	    if (stats_arr[i].count)
		fprintf(out, "%lg %lg\n", (i + 0.5)*PLOT_NODECOUNT_GRAN,
			stats_arr[i].time
			/((double)CLOCKS_PER_SEC*stats_arr[i].count));
	fclose(out);
    }

    if (cu_test_bug_count()) {
	cu_errf("%d tests failed.", cu_test_bug_count());
	return 2;
    }
    return 0;
}
