/*-*-C-*-
 * Copyright 2006  Petter Urkedal
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cuflow/workers.h>
#include <cucon/priq.h>
#include <cucon/list.h>
#include <cu/thread.h>
#include <cu/memory.h>
#include <errno.h>
#include <atomic_ops.h>

typedef struct cuflow_worker_s *cuflow_worker_t;
typedef struct cuflow_atjob_s *cuflow_atjob_t;

struct cuflow_atjob_s
{
    struct timespec t_call;
    cu_clop(callback, void, struct timespec *current_time);
};
struct cuflow_worker_s
{
    pthread_t thread;
    cu_bool_t do_shutdown;
};

static pthread_mutex_t		work_mutex = CU_MUTEX_INITIALISER;
static pthread_cond_t		work_cond = PTHREAD_COND_INITIALIZER;
static struct cucon_priq_s	work_atq;
static struct cucon_list_s	work_nowlist;

static pthread_mutex_t		workers_mutex = CU_MUTEX_INITIALISER;
static struct cucon_list_s	workers_list; /* of struct cucon_worker_s */
static AO_t			workers_count = 0;


CU_SINLINE cu_bool_t
timespec_before(struct timespec *t0, struct timespec *t1)
{
    if (t0->tv_sec < t1->tv_sec)
	return cu_true;
    if (t0->tv_sec > t1->tv_sec)
	return cu_false;
    return t0->tv_nsec < t1->tv_nsec;
}

cu_clop_def(atjob_before_clop, cu_bool_t, void *atjob0, void *atjob1)
{
    return timespec_before(&((cuflow_atjob_t)atjob0)->t_call,
			   &((cuflow_atjob_t)atjob1)->t_call);
}

static void
worker_atrun(cuflow_worker_t worker)
{
    struct timespec now;
    clock_gettime(CLOCK_REALTIME, &now);
    while (!cucon_priq_is_empty(&work_atq)) {
	cuflow_atjob_t atjob = cucon_priq_front(&work_atq);
	cuflow_atjob_t atjob_next;
	if (timespec_before(&now, &atjob->t_call))
	    break;
	cucon_priq_pop_front(&work_atq);

	/* Signal more worker threads in case there is more to do right
	 * away. */
	atjob_next = cucon_priq_front(&work_atq);
	if (timespec_before(&now, &atjob_next->t_call))
	    pthread_cond_signal(&work_cond);

	cu_mutex_unlock(&work_mutex);
	cu_call(atjob->callback, &now);
	cu_mutex_lock(&work_mutex);
    }
}

static void *
worker_main(void *worker)
{
#define worker ((cuflow_worker_t)worker)
    cu_mutex_lock(&work_mutex);
    while (!worker->do_shutdown) {
	while (!cucon_list_is_empty(&work_nowlist)) {
	    cu_clop0(callback, void);
	    callback = cucon_list_front_ptr(&work_nowlist);
	    cucon_list_pop_front(&work_nowlist);
	    cu_mutex_unlock(&work_mutex);
	    cu_call0(callback);
	    cu_mutex_lock(&work_mutex);
	    if (worker->do_shutdown)
		goto break_0;
	}
	if (cucon_priq_is_empty(&work_atq)) {
	    pthread_cond_wait(&work_cond, &work_mutex);
	}
	else {
	    cuflow_atjob_t atjob = cucon_priq_front(&work_atq);
	    int st;
	    st = pthread_cond_timedwait(&work_cond, &work_mutex,
					&atjob->t_call);
	    if (st == ETIMEDOUT) {
		if (worker->do_shutdown)
		    break;
		worker_atrun(worker);
	    }
	}
    }
break_0:
    pthread_cond_signal(&work_cond);
    cu_mutex_unlock(&work_mutex);
    return NULL;
#undef worker
}

int
cuflow_workers_count()
{
    return AO_load_acquire_read(&workers_count);
}

static void
workers_drop_lck(int target_count)
{
    do {
	int last_workers_count;
	cuflow_worker_t worker;

	cu_debug_assert(!cucon_list_is_empty(&workers_list));
	worker = cucon_list_front_mem(&workers_list);
	cucon_list_pop_front(&workers_list);
	worker->do_shutdown = cu_true;
	--workers_count;
	last_workers_count = workers_count;

	cu_mutex_unlock(&workers_mutex);
	cu_mutex_lock(&work_mutex);
	pthread_cond_broadcast(&work_cond);
	cu_mutex_unlock(&work_mutex);
	pthread_join(worker->thread, NULL);
	cu_mutex_lock(&workers_mutex);

	/* If strict monononicity of workers_count is broken, it means
	 * that another thread increased the number of workers while
	 * workers_mutex was unlocked.  If so, give up. */
	if (workers_count > last_workers_count)
	    break; /* Another thread is trying to spawn new workers. */
    } while (target_count < workers_count);
}

static void
workers_spawn_lck(int target_count)
{
    while (target_count > workers_count) {
	cucon_listnode_t worker_it;
	cuflow_worker_t worker;

	worker_it = cucon_list_append_mem(&workers_list,
					  sizeof(struct cuflow_worker_s));
	worker = cucon_listnode_mem(worker_it);
	worker->do_shutdown = cu_false;
	cu_pthread_create(&worker->thread, NULL, worker_main, worker);
	++workers_count;
    }
}

void
cuflow_workers_spawn(int target_count)
{
    cu_mutex_lock(&workers_mutex);
    if (target_count < workers_count)
	workers_drop_lck(target_count);
    else if (target_count > workers_count)
	workers_spawn_lck(target_count);
    cu_mutex_unlock(&workers_mutex);
}

void
cuflow_workers_spawn_at_most(int target_count)
{
    cu_mutex_lock(&workers_mutex);
    if (target_count < workers_count)
	workers_drop_lck(target_count);
    cu_mutex_unlock(&workers_mutex);
}

void
cuflow_workers_spawn_at_least(int target_count)
{
    cu_mutex_lock(&workers_mutex);
    if (target_count > workers_count)
	workers_spawn_lck(target_count);
    cu_mutex_unlock(&workers_mutex);
}

void
cuflow_workers_call_at(cu_clop(callback, void, struct timespec *t_now),
		       struct timespec *t_call)
{
    cuflow_atjob_t atjob = cu_gnew(struct cuflow_atjob_s);
    cuflow_atjob_t atjob_prev_head;

    memcpy(&atjob->t_call, t_call, sizeof(struct timespec));
    atjob->callback = callback;

    cu_mutex_lock(&work_mutex);
    if (cucon_priq_is_empty(&work_atq))
	atjob_prev_head == NULL;
    else
	atjob_prev_head = cucon_priq_front(&work_atq);
    cucon_priq_insert(&work_atq, atjob);

    /* Make sure we have at least one worker which will do the timed job.
     * This worker will signal (from worker_atrun) another worker if there is
     * more work to do right away. */
    if (!atjob_prev_head || timespec_before(t_call, &atjob_prev_head->t_call))
	pthread_cond_signal(&work_cond);

    cu_mutex_unlock(&work_mutex);
}

void
cuflow_workers_call(cu_clop0(callback, void))
{
    cu_mutex_lock(&work_mutex);
    cucon_list_append_ptr(&work_nowlist, (void *)callback);
    pthread_cond_signal(&work_cond);
    cu_mutex_unlock(&work_mutex);
}

static void
workers_cleanup(void)
{
    cuflow_workers_spawn(0);
}

void
cuflowP_workers_init()
{
    cucon_priq_cct(&work_atq, atjob_before_clop);
    cucon_list_cct(&work_nowlist);
    cucon_list_cct(&workers_list);
    atexit(workers_cleanup);
}
